from drakaina.contrib.django.middleware import JWTAuthenticationMiddleware
from drakaina.contrib.django.views import RPCView
from drakaina.contrib.jwt import SUPPORTED_ALGORITHMS

__all__ = (
    "JWTAuthenticationMiddleware",
    "RPCView",
    "CREDENTIALS_REQUIRED",
    "JWT_ALGORITHMS",
    "JWT_SECRET_KEY",
    "JWT_PUBLIC_KEY",
    "JWT_PREFIX",
    "JWT_USE_COOKIES",
    "JWT_COOKIE_KEY",
    "JWT_DECODE_OPTIONS",
    "JWT_VERIFY_VALUES",
    "JWT_LEEWAY",
    "JWT_USER_ID_FIELD",
    "JWT_TOKEN_GETTER",
    "JWT_USER_GETTER",
    "JWT_SCOPES_GETTER",
    "JWT_REVOKE_CHECKER",
)


# Settings default values
# ❗ : In settings.py your django project use `DRAKAINA_` prefix

CREDENTIALS_REQUIRED = True
JWT_ALGORITHMS = SUPPORTED_ALGORITHMS
JWT_SECRET_KEY = None
JWT_PUBLIC_KEY = None
JWT_PREFIX = "Bearer"
JWT_USE_COOKIES = False
JWT_COOKIE_KEY = "jwt"
JWT_DECODE_OPTIONS = {
    "require": [],
    "verify_iss": False,
    "verify_aud": False,
    "verify_exp": True,
    "verify_nbf": False,
    "verify_iat": True,
    "verify_jti": False,
}
JWT_VERIFY_VALUES = {}
JWT_LEEWAY = 0
JWT_USER_ID_FIELD = "user_id"
JWT_TOKEN_GETTER = None
JWT_USER_GETTER = None
JWT_SCOPES_GETTER = None
JWT_REVOKE_CHECKER = None



import asyncio
import json
import logging
import typing

from aiohttp import ClientSession, http_websocket, web_ws

from .base import BaseJsonRpcClient
from .. import errors, typedefs, utils


__all__ = (
    'WsJsonRpcClient',
)

logger = logging.getLogger(__name__)


class WsJsonRpcClient(BaseJsonRpcClient):
    url: typing.Optional[str]
    ws_connect: typing.Optional[typedefs.WSConnectType]
    ws_connect_kwargs: dict
    timeout: typing.Optional[int]
    _pending: typing.Dict[typing.Any, asyncio.Future]
    _message_worker: typing.Optional[asyncio.Future] = None
    _session_is_outer: bool
    _ws_connect_is_outer: bool
    _json_request_handler: typing.Optional[typing.Callable] = None
    _unprocessed_json_response_handler: typing.Optional[typing.Callable] = None
    _background_tasks: typing.Set

    def __init__(self,
                 url: typing.Optional[str] = None, *,
                 session: typing.Optional[ClientSession] = None,
                 ws_connect: typing.Optional[typedefs.WSConnectType] = None,
                 timeout: typing.Optional[int] = 5,
                 json_request_handler: typing.Optional[typing.Callable] = None,
                 unprocessed_json_response_handler: typing.Optional[typing.Callable] = None,
                 **ws_connect_kwargs) -> None:
        assert (session is not None) or (url is not None and session is None) or (ws_connect is not None)

        self.url = url
        self.timeout = timeout

        self.session = session
        self._session_is_outer = session is not None  # We don't close an outer session.

        self.ws_connect = ws_connect
        self.ws_connect_kwargs = ws_connect_kwargs
        self._ws_connect_is_outer = ws_connect is not None  # We don't close an outer WS connection.

        self._pending = {}
        self._json_request_handler = json_request_handler
        self._unprocessed_json_response_handler = unprocessed_json_response_handler
        self._background_tasks = set()

    async def connect(self) -> None:
        if self.session is None and self.ws_connect is None:
            self.session = ClientSession(json_serialize=self.json_serialize)

        if self.ws_connect is None:
            assert self.url is not None and self.session is not None

            try:
                self.ws_connect = await self.session.ws_connect(self.url, **self.ws_connect_kwargs)
            except Exception:
                await self.disconnect()
                raise

        self._message_worker = asyncio.create_task(self._handle_ws_messages())

    async def disconnect(self) -> None:
        if self.ws_connect is not None and not self._ws_connect_is_outer:
            await self.ws_connect.close()

        if self.session is not None and not self._session_is_outer:
            await self.session.close()

        if self._message_worker is not None:
            await self._message_worker

    async def send_json(self,
                        data: typing.Any, *,
                        without_response: bool = False,
                        **kwargs) -> typing.Tuple[typing.Any, typing.Optional[dict]]:
        assert self.ws_connect is not None

        if without_response:
            await self.ws_connect.send_str(self.json_serialize(data), **kwargs)
            return None, None

        request_ids = self._get_ids_from_json(data)
        future: asyncio.Future = asyncio.Future()

        for request_id in request_ids:
            self._pending[request_id] = future

        await self.ws_connect.send_str(self.json_serialize(data), **kwargs)

        if not request_ids:
            return None, None

        if self.timeout is not None:
            future = asyncio.wait_for(future, timeout=self.timeout)  # type: ignore

        result = await future

        return result, None

    def clear_pending(self) -> None:
        self._pending.clear()

    @staticmethod
    def _get_ids_from_json(data: typing.Any) -> typing.Tuple[typedefs.JsonRpcIdType, ...]:
        if not data:
            return ()

        if isinstance(data, typing.Mapping) and data.get('id') is not None:
            return (
                data['id'],
            )

        if isinstance(data, typing.Sequence):
            return tuple(
                item['id']
                for item in data
                if isinstance(item, typing.Mapping) and item.get('id') is not None
            )

        return ()

    async def _handle_ws_messages(self) -> None:
        assert self.ws_connect is not None

        ws_msg: http_websocket.WSMessage

        async for ws_msg in self.ws_connect:
            if ws_msg.type != http_websocket.WSMsgType.TEXT:
                continue

            try:
                task = asyncio.create_task(self._handle_single_ws_message(ws_msg))
            except asyncio.CancelledError as e:
                error = errors.ServerError(utils.get_exc_message(e)).with_traceback()
                self._notify_all_about_error(error)
                raise
            except Exception:
                logger.warning('Can not process WS message.', exc_info=True)
            else:
                # To avoid a task disappearing mid execution:
                self._background_tasks.add(task)
                task.add_done_callback(self._background_tasks.discard)

    async def _handle_single_ws_message(self, ws_msg: http_websocket.WSMessage) -> None:
        if ws_msg.type != http_websocket.WSMsgType.text:
            return

        try:
            json_response = json.loads(ws_msg.data)
        except Exception:
            logger.warning('Cann\'t parse json.', exc_info=True)
            return

        if not json_response:
            return

        if isinstance(json_response, typing.Mapping):
            await self._handle_single_json_response(json_response, ws_msg=ws_msg)
            return

        if isinstance(json_response, typing.Sequence):
            await self._handle_json_responses(json_response, ws_msg=ws_msg)
            return

        logger.warning('Couldn\'t process the response.', extra={
            'json_response': json_response,
        })

    async def _handle_single_json_response(self, json_response: typing.Mapping, *, ws_msg: web_ws.WSMessage) -> None:
        if 'method' in json_response:
            if self._json_request_handler is not None:
                await self._json_request_handler(
                    ws_connect=self.ws_connect,
                    ws_msg=ws_msg,
                    json_request=json_response,
                )
        elif 'id' in json_response and json_response['id'] in self._pending:
            self._notify_about_result(json_response['id'], json_response)
        elif self._unprocessed_json_response_handler is not None:
            self._unprocessed_json_response_handler(
                ws_connect=self.ws_connect,
                ws_msg=ws_msg,
                json_response=json_response,
            )

    async def _handle_json_responses(self, json_responses: typing.Sequence, *, ws_msg: web_ws.WSMessage) -> None:
        if isinstance(json_responses[0], typing.Mapping) and 'method' in json_responses[0]:
            if self._json_request_handler is not None:
                await self._json_request_handler(ws_connect=self.ws_connect, ws_msg=ws_msg)
        else:
            response_ids = self._get_ids_from_json(json_responses)

            if response_ids:
                self._notify_about_results(response_ids, json_responses)
            elif self._unprocessed_json_response_handler is not None:
                self._unprocessed_json_response_handler(
                    ws_connect=self.ws_connect,
                    ws_msg=ws_msg,
                    json_response=json_responses,
                )

    def _notify_all_about_error(self, error: errors.JsonRpcError) -> None:
        for future in self._pending.values():
            future.set_exception(error)

        self.clear_pending()

    def _notify_about_result(self, response_id: typedefs.JsonRpcIdType, json_response: typing.Mapping) -> None:
        future = self._pending.pop(response_id, None)

        if future is not None:
            future.set_result(json_response)

    def _notify_about_results(self,
                              response_ids: typing.Sequence[typedefs.JsonRpcIdType],
                              json_response: typing.Sequence) -> None:
        is_processed = False

        for response_id in response_ids:
            future = self._pending.pop(response_id, None)

            if future is not None and not is_processed:
                # We suppose that a batch result has the same ids that we sent.
                # And this ids have the same future.

                future.set_result(json_response)
                is_processed = True



import logging

from drakaina.middleware.base import BaseMiddleware


class LoggingMiddleware(BaseMiddleware):
    logger: object

    async def logging_middleware(self, request, handler):
        raw_request = request.dump()

        self.logger.info(
            'RpcRequest id="%s" method="%s" params="%s"',
            raw_request.get("id", ""),
            raw_request["method"],
            raw_request.get("params", ""),
            extra={"request": raw_request},
        )

        response = await handler(request)

        raw_response = request.dump()

        self.logger.info(
            'RpcResponse id="%s" method="%s" params="%s" result="%s" error="%s"',
            raw_request.get("id", ""),
            raw_request["method"],
            raw_request.get("params", ""),
            raw_response.get("result", ""),
            raw_response.get("error", ""),
            extra={"request": raw_response, "response": raw_response},
        )

        return response



import abc
import types
import typing
from functools import partial

from .. import errors, protocol, typedefs, utils


__all__ = (
    'BaseJsonRpcClient',
)


class BaseJsonRpcClient(abc.ABC):
    error_map: typing.Mapping[int, typing.Type[errors.JsonRpcError]] = {
        error.code: error
        for error in errors.DEFAULT_KNOWN_ERRORS
    }
    json_serialize: typing.Callable = utils.json_serialize

    async def __aenter__(self) -> 'BaseJsonRpcClient':
        await self.connect()
        return self

    async def __aexit__(self,
                        exc_type: typing.Optional[typing.Type[BaseException]],
                        exc_value: typing.Optional[BaseException],
                        traceback: typing.Optional[types.TracebackType]) -> None:
        await self.disconnect()

    def __getattr__(self, method_name: str) -> typing.Callable:
        return partial(self.call, method_name)

    @abc.abstractmethod
    async def connect(self) -> None:
        pass

    @abc.abstractmethod
    async def disconnect(self) -> None:
        pass

    async def call(self, method_name: str, *args, **kwargs) -> typing.Any:
        request = protocol.JsonRpcRequest(id=utils.get_random_id(), method_name=method_name, args=args, kwargs=kwargs)
        response = await self.direct_call(request)

        assert response is not None  # Because it isn't a notification

        if response.error is not None:
            raise response.error

        return response.result

    async def notify(self, method_name: str, *args, **kwargs) -> None:
        request = protocol.JsonRpcRequest(method_name=method_name, args=args, kwargs=kwargs)
        await self.direct_call(request)

    async def batch(self,
                    method_descriptions: typedefs.ClientMethodDescriptionsType, *,
                    save_order: bool = True) -> typing.Sequence:
        if isinstance(method_descriptions, protocol.JsonRpcBatchRequest):
            batch_request = method_descriptions
        else:
            batch_request = protocol.JsonRpcBatchRequest(requests=tuple(
                self._parse_method_description(method_description)
                for method_description in method_descriptions
            ))

        batch_response = await self.direct_batch(batch_request)

        assert batch_response is not None  # Because it isn't a notification

        if save_order:
            return utils.collect_batch_result(batch_request, batch_response)
        else:
            return tuple(
                response.result if response.error is None else response.error
                for response in batch_response.responses
            )

    async def batch_notify(self, method_descriptions: typedefs.ClientMethodDescriptionsType) -> None:
        if isinstance(method_descriptions, protocol.JsonRpcBatchRequest):
            batch_request = method_descriptions
        else:
            batch_request = protocol.JsonRpcBatchRequest(requests=tuple(
                self._parse_method_description(method_description, is_notification=True)
                for method_description in method_descriptions
            ))

        await self.direct_batch(batch_request)

    async def direct_call(self,
                          request: protocol.JsonRpcRequest,
                          **kwargs) -> typing.Optional[protocol.JsonRpcResponse]:
        json_response, context = await self.send_json(
            request.dump(),
            without_response=request.is_notification,
            **kwargs,
        )

        if request.is_notification:
            return None

        response = protocol.JsonRpcResponse.load(
            json_response,
            error_map=self.error_map,
            context=context,
        )

        return response

    async def direct_batch(self,
                           batch_request: protocol.JsonRpcBatchRequest,
                           **kwargs) -> typing.Optional[protocol.JsonRpcBatchResponse]:
        if not batch_request.requests:
            raise errors.InvalidRequest('You can not send an empty batch request.')

        is_notification = batch_request.is_notification

        json_response, context = await self.send_json(
            batch_request.dump(),
            without_response=is_notification,
            **kwargs,
        )

        if is_notification:
            return None

        if not json_response:
            raise errors.ParseError('Server returned an empty batch response.')

        return protocol.JsonRpcBatchResponse.load(json_response)

    @abc.abstractmethod
    async def send_json(self,
                        data: typing.Any, *,
                        without_response: bool = False,
                        **kwargs) -> typing.Tuple[typing.Any, typing.Optional[dict]]:
        pass

    @staticmethod
    def _parse_method_description(method_description: typedefs.ClientMethodDescriptionType, *,
                                  is_notification: bool = False) -> protocol.JsonRpcRequest:
        if isinstance(method_description, protocol.JsonRpcRequest):
            return method_description

        request_id = None if is_notification else utils.get_random_id()

        if isinstance(method_description, str):
            return protocol.JsonRpcRequest(
                id=request_id,
                method_name=method_description,
            )

        if len(method_description) == 1:
            return protocol.JsonRpcRequest(
                id=request_id,
                method_name=method_description[0],
            )

        if len(method_description) == 2:
            return protocol.JsonRpcRequest(
                id=request_id,
                method_name=method_description[0],
                params=method_description[1],
            )

        if len(method_description) == 3:
            return protocol.JsonRpcRequest(
                id=request_id,
                method_name=method_description[0],
                args=method_description[1],
                kwargs=method_description[2],  # type: ignore
            )

        raise errors.InvalidParams('Use string or list (length less than or equal to 3).')



from drakaina.exceptions import InternalServerError
from drakaina.exceptions import InvalidTokenError


class JWTBackendError(InternalServerError):
    """JWT backend error"""


class InvalidJWTTokenError(InvalidTokenError):
    """Invalid JWT token error"""


class ValidationJWTTokenError(InvalidTokenError):
    """JWT token validation error"""



from typing import Dict

from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.contrib.auth.models import update_last_login
from django.http import HttpRequest

from drakaina import remote_procedure
from drakaina.exceptions import AuthenticationFailedError
from drakaina.contrib.jwt.utils import decode_jwt_token
from drakaina.contrib.jwt.utils import get_tokens_for
from drakaina.contrib.jwt.utils import refresh_tokens

UserModel = get_user_model()
UPDATE_LAST_LOGIN = True
ROTATE_REFRESH_TOKENS = False


def get_user(request: HttpRequest, username: str, password: str) -> UserModel:
    authenticate_kwargs = {
        UserModel.USERNAME_FIELD: username,
        "password": password,
        "request": request,
    }
    user = authenticate(**authenticate_kwargs)

    if user is None or not user.is_active:
        raise AuthenticationFailedError(
            "No active account found with the given credentials",
        )

    return user


@remote_procedure(name="token", provide_request=True)
def get_token(
    request: HttpRequest,
    username: str,
    password: str,
) -> Dict[str, str]:
    user = get_user(request, username, password)

    tokens = get_tokens_for(user.id)

    if UPDATE_LAST_LOGIN:
        update_last_login(None, user)

    return tokens


@remote_procedure(name="token.refresh")
def refresh_token(refresh: str) -> Dict[str, str]:
    return refresh_tokens(
        refresh,
        verify=True,
        rotate_refresh=ROTATE_REFRESH_TOKENS,
    )


@remote_procedure(name="token.verify")
def verify_token(token: str) -> bool:
    decode_jwt_token(token, [], "", "")
    return True



from __future__ import annotations

from asyncio import iscoroutinefunction
from functools import update_wrapper
from typing import Any
from typing import Callable
from typing import Iterable
from typing import MutableMapping
from typing import Optional
from typing import TypeVar

from drakaina._types import AnyRequest
from drakaina._types import Comparator
from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import ForbiddenError
from drakaina.registries import RPC_REGISTRY
from drakaina.registries import RPCRegistry
from drakaina.utils import iterable_str_arg

__all__ = (
    "RPCRegistry",
    "rpc_registry",
    "remote_procedure",
    "check_permissions",
    "login_required",
    "match_all",
    "match_any",
    "ENV_APP",
    "ENV_IS_AUTHENTICATED",
    "ENV_USER",
    "ENV_USER_ID",
    "ENV_AUTH_PAYLOAD",
    "ENV_AUTH_SCOPES",
)

# General registry
rpc_registry = RPCRegistry()

# Environ (Request) context
ENV_APP = "drakaina.app"
ENV_IS_AUTHENTICATED = "auth.is_authenticated"
ENV_USER = "user"
ENV_USER_ID = "user_id"
ENV_AUTH_PAYLOAD = "auth.payload"
ENV_AUTH_SCOPES = "auth.scopes"

T = TypeVar("T")


def remote_procedure(
    name: Optional[str] = None,
    registry: Optional[RPCRegistry] = None,
    provide_request: Optional[bool] = None,
    metadata: Optional[dict[str, Any]] = None,
    **meta_options,
) -> Callable:
    """Decorator allow wrap function and define it as remote procedure.

    :param name:
        Procedure name. Default as function name.
    :type name: str
    :param registry:
        Procedure registry custom object
    :type registry: RPCRegistry
    :param provide_request:
        Provide a request object or context data (from the transport layer).
        If `True`, then the request object or context can be supplied to
        the procedure as a `request` argument.
    :type provide_request: bool
    :param metadata:
        Metadata that can be processed by middleware.
    :type metadata: dict[str, Any]
    :param meta_options:
        Metadata that can be processed by middleware.

    """

    def __decorator(
        procedure: T,
        _registry: RPCRegistry = None,
        _name: str = None,
        _provide_request: bool = None,
        **_meta_options,
    ) -> T:
        """Returns a registered procedure"""

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                if not _provide_request:
                    if len(args) == 0:
                        kwargs.pop("request")
                    else:
                        scope, *args = args
                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                if not _provide_request:
                    if len(args) == 0:
                        kwargs.pop("request")
                    else:
                        environ, *args = args
                return procedure(*args, **kwargs)

        # Need to update the wrapper before registering in the registry
        decorated_procedure = update_wrapper(wrapper, procedure)

        if _registry is None:
            _registry = rpc_registry
        if _name is None:
            _name = procedure.__name__
        _metadata = _meta_options.pop("metadata") or {}
        _registry.register_procedure(
            decorated_procedure,
            name=_name,
            provide_request=_provide_request,
            metadata={**_metadata, **_meta_options},
        )

        return decorated_procedure

    if callable(name):
        return __decorator(
            name,
            registry,
            None,
            provide_request,
            metadata=metadata,
            **meta_options,
        )
    elif not isinstance(name, (str, type(None))):
        raise TypeError(
            "Expected first argument to be an str, a callable, or None",
        )

    def decorator(procedure):
        assert callable(procedure)
        return __decorator(
            procedure,
            registry,
            name,
            provide_request,
            metadata=metadata,
            **meta_options,
        )

    return decorator


def login_required(*_args) -> Callable:
    """Requires login decorator.

    Gives access to the procedure only to authenticated users.

    """

    def __decorator(procedure: T) -> T:
        """Returns a registered procedure"""
        registry: RPCRegistry = getattr(procedure, RPC_REGISTRY, None)
        if registry is None:
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)
                if not _is_authenticated(request):
                    raise ForbiddenError("Authentication required")

                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)
                if not _is_authenticated(request):
                    raise ForbiddenError("Authentication required")

                return procedure(*args, **kwargs)

        wrapper = update_wrapper(wrapper, procedure)
        registry.replace(procedure, wrapper)
        return wrapper

    if len(_args) > 0:
        if callable(_args[0]):
            return __decorator(_args[0])
        else:
            raise TypeError("Expected first argument to be a callable")

    def decorator(procedure):
        assert callable(procedure)
        return __decorator(procedure)

    return decorator


def match_any(required: Iterable[str], provided: str | Iterable[str]) -> bool:
    return any((scope in provided for scope in required))


def match_all(required: Iterable[str], provided: str | Iterable[str]) -> bool:
    return set(required).issubset(set(provided))


def check_permissions(
    scopes: str | Iterable[str],
    comparator: Comparator = match_all,
) -> Callable:
    """Permission decorator.

    Gives access to the procedure only to authorized users.

    """
    if not (
        isinstance(scopes, str)
        or (
            isinstance(scopes, Iterable)
            and all((isinstance(scope, str) for scope in scopes))
        )
    ):
        raise TypeError(
            "The `scopes` argument must be a string or Iterable[string]",
        )
    if not callable(comparator):
        raise TypeError(
            "The `comparator` argument must be a function that implements "
            "the Comparator interface",
        )

    procedure_scopes = iterable_str_arg(scopes)

    def __decorator(procedure: T) -> T:
        registry: RPCRegistry = getattr(procedure, RPC_REGISTRY, None)
        if registry is None:
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)  # noqa

                if not _is_authenticated(request):
                    raise ForbiddenError("Authentication required")

                user_scopes = _get_scopes(request)
                if not isinstance(user_scopes, Iterable):
                    raise AuthenticationFailedError(
                        "Invalid permissions format",
                    )

                if not comparator(procedure_scopes, user_scopes):
                    raise ForbiddenError("Forbidden")

                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)  # noqa

                if not _is_authenticated(request):
                    raise ForbiddenError("Authentication required")

                user_scopes = _get_scopes(request)
                if not isinstance(user_scopes, Iterable):
                    raise AuthenticationFailedError(
                        "Invalid permissions format",
                    )

                if not comparator(procedure_scopes, user_scopes):
                    raise ForbiddenError("Forbidden")

                return procedure(*args, **kwargs)

        wrapper = update_wrapper(wrapper, procedure)
        registry.replace(procedure, wrapper)
        return wrapper

    return __decorator


def _get_request(*args, **kwargs) -> Optional[AnyRequest]:
    if len(args) == 0:
        return kwargs.get("request")
    else:
        return args[0]


def _is_authenticated(request: AnyRequest) -> bool:
    if isinstance(request, (dict, MutableMapping)):
        return request.get(ENV_IS_AUTHENTICATED, False)
    else:
        return getattr(request, ENV_IS_AUTHENTICATED, False)


def _get_scopes(request: AnyRequest) -> Optional[Iterable[str]]:
    if isinstance(request, (dict, MutableMapping)):
        scopes = request.get(ENV_AUTH_SCOPES, None)
    else:
        scopes = getattr(request, ENV_AUTH_SCOPES, None)
    if scopes is not None:
        return iterable_str_arg(scopes)



from __future__ import annotations

from typing import Any
from typing import Iterable
from typing import Optional
from typing import Protocol

from drakaina._types import ASGIScope
from drakaina._types import WSGIEnvironment


class TokenGetter(Protocol):
    def __call__(
        self,
        request: ASGIScope | WSGIEnvironment,
    ) -> Optional[str]:
        ...


class ScopesGetter(Protocol):
    def __call__(
        self,
        request: ASGIScope | WSGIEnvironment,
        payload: dict[str, Any],
    ) -> Optional[str, Iterable[str]]:
        ...


class RevokeChecker(Protocol):
    def __call__(
        self,
        request: ASGIScope | WSGIEnvironment,
        payload: dict[str, Any],
    ) -> bool:
        ...


class UserGetter(Protocol):
    def __call__(
        self,
        request: ASGIScope | WSGIEnvironment,
        payload: dict[str, Any],
    ) -> Optional[Any]:
        ...



"""
https://github.com/lidatong/dataclasses-json/
https://github.com/Fatal1ty/mashumaro

https://github.com/horejsek/python-fastjsonschema
https://github.com/s-knibbs/dataclasses-jsonschema
https://github.com/marshmallow-code/apispec
https://github.com/strongbugman/apiman

starlette.starlette.schemas.BaseSchemaGenerator
"""
from inspect import Parameter
from typing import Any
from typing import Dict
from typing import ForwardRef


def get_typed_annotation(param: Parameter, globalns: Dict[str, Any]) -> Any:
    annotation = param.annotation
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        # annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


# TODO: move to openapi middleware
# _signature = signature(procedure)
# global_ns = getattr(procedure, "__globals__", {})
# typed_params = [
#     Parameter(
#         name=param.name,
#         kind=param.kind,
#         default=param.default,
#         annotation=get_typed_annotation(param, global_ns),
#     )
#     for param in _signature.parameters.values()
# ]
# typed_signature = Signature(typed_params)



from drakaina._types import ASGIReceive
from drakaina._types import ASGIScope
from drakaina._types import ASGISend
from drakaina._types import ProxyRequest
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse
from drakaina.middleware.base import BaseMiddleware


class RequestWrapperMiddleware(BaseMiddleware):
    """The middleware for wrapping the request object.

    Provides access to the mapping environment object through
    the attribute access interface. This is needed for some
    backward compatibility in cases where the request is
    an object with attributes, such as `request.user`.

    """

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        return self.app(ProxyRequest(environ), start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        await self.app(ProxyRequest(scope), receive, send)



from drakaina.middleware.base import BaseMiddleware
from drakaina.middleware.cors import CORSMiddleware
from drakaina.middleware.exception import ExceptionMiddleware
from drakaina.middleware.request_wrapper import RequestWrapperMiddleware

__all__ = (
    "BaseMiddleware",
    "CORSMiddleware",
    "ExceptionMiddleware",
    "RequestWrapperMiddleware",
)



from __future__ import annotations

from functools import partial
from logging import Logger
from typing import Iterable
from typing import Optional
from typing import Type
from typing import Union

from drakaina import ENV_APP
from drakaina._types import WSGIApplication
from drakaina._types import WSGIEnvironment
from drakaina._types import WSGIInputStream
from drakaina._types import WSGIResponse
from drakaina._types import WSGIStartResponse
from drakaina.exceptions import BadRequestError
from drakaina.middleware.base import BaseMiddleware
from drakaina.middleware.exception import ExceptionMiddleware
from drakaina.middleware.request_wrapper import RequestWrapperMiddleware
from drakaina.rpc_protocols import BaseRPCProtocol
from drakaina.rpc_protocols import JsonRPCv2

ALLOWED_METHODS = ("OPTIONS", "GET", "POST")
Middlewares = Iterable[Type[BaseMiddleware] | partial[BaseMiddleware]]


class WSGIHandler:
    """Implementation of WSGI protocol.

    :param route:
    :type route: str
    :param handler: RPC protocol implementation.
    :type handler: BaseRPCProtocol
    :param middlewares: List of WSGI middlewares.
    :type middlewares: Iterable[Type[BaseMiddleware] | partial[BaseMiddleware]]
    :param logger: A `logging.Logger` object.
    :type logger: Logger
    :param max_content_size: Limiting request body size for DoS protection.
    :type max_content_size: int
    :param provide_smd:
    :type provide_smd: bool | str
    :param provide_openrpc:
    :type provide_openrpc: bool | str
    :param provide_openapi:
    :type provide_openapi: bool | str

    """

    __slots__ = (
        "environ",
        "start_response",
        "handler",
        "route",
        "logger",
        "max_content_size",
        "provide_smd",
        "provide_openrpc",
        "provide_openapi",
        "_rpc_content_type",
        "_allowed_methods",
        "_middlewares_chain",
    )

    environ: WSGIEnvironment
    start_response: WSGIStartResponse
    _middlewares_chain: Union[WSGIApplication, BaseMiddleware]

    def __init__(
        self,
        route: Optional[str] = None,
        handler: Optional[BaseRPCProtocol] = None,
        middlewares: Optional[Middlewares] = None,
        logger: Optional[Logger] = None,
        max_content_size: int = 4096,
        provide_smd: Optional[Union[bool, str]] = False,
        provide_openrpc: Optional[Union[bool, str]] = False,
        provide_openapi: Optional[Union[bool, str]] = False,
    ):
        self.handler = handler if handler is not None else JsonRPCv2()
        self._rpc_content_type = self.handler.content_type

        self.route = route
        if isinstance(self.route, str) and not self.route.startswith("/"):
            self.route = "/" + self.route

        self.logger = logger
        self.max_content_size = int(max_content_size)
        self.provide_smd = provide_smd
        self.provide_openrpc = provide_openrpc
        self.provide_openapi = provide_openapi

        if provide_smd or provide_openrpc or provide_openapi:
            self._allowed_methods = ", ".join(ALLOWED_METHODS)
        else:
            self._allowed_methods = ", ".join(
                [m for m in ALLOWED_METHODS if m != "GET"],
            )

        # Build middleware stack
        self._middlewares_chain = self._wsgi_app
        kw = {"is_async": False}
        for mw in reversed(middlewares or []):
            if (
                isinstance(mw, partial)
                and issubclass(mw.func, BaseMiddleware)
                or issubclass(mw, BaseMiddleware)
            ):
                self._middlewares_chain = mw(self._middlewares_chain, **kw)
            self._middlewares_chain = mw(self._middlewares_chain)

        # The middleware for handling exceptions in the middleware according
        #  to the RPC protocol.
        self._middlewares_chain = ExceptionMiddleware(
            RequestWrapperMiddleware(self._middlewares_chain, **kw),  # noqa
            handler=self.handler,
            logger=self.logger,
            **kw,
        )

    def __call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        environ[ENV_APP] = self
        return self._middlewares_chain(environ, start_response)  # noqa

    def _wsgi_app(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        self.environ = environ
        self.start_response = start_response

        method = environ["REQUEST_METHOD"]
        if method in ALLOWED_METHODS:
            if self._validate_path(environ["PATH_INFO"]):
                return getattr(self, method.lower())()

            return self._not_found()

        return self._method_not_allowed()

    def get(self) -> WSGIResponse:
        if self.provide_smd:
            response_body = self.handler.smd_scheme()
            response_headers = [
                ("Content-Type", "application/json"),
                ("Content-Length", str(len(response_body))),
            ]
        elif self.provide_openrpc:
            response_body = self.handler.openrpc_scheme()
            response_headers = [
                ("Content-Type", "application/json"),
                ("Content-Length", str(len(response_body))),
            ]
        elif self.provide_openapi:
            response_body = self.handler.openapi_scheme()
            response_headers = [
                ("Content-Type", "application/json"),
                ("Content-Length", str(len(response_body))),
            ]
        else:
            return self._method_not_allowed()

        self.start_response("200 OK", response_headers)
        return (response_body,)

    def post(self) -> WSGIResponse:
        wsgi_input: WSGIInputStream = self.environ["wsgi.input"]

        content_type = self.environ.get("CONTENT_TYPE")
        content_length = int(self.environ.get("CONTENT_LENGTH") or 0)
        if (
            not (content_type and content_length)
            or content_type != self._rpc_content_type
            or content_length > self.max_content_size
        ):
            if content_type != self._rpc_content_type:
                response_status = "415 Unsupported Media Type"
            else:
                response_status = "400 Bad Request"
            # Return RPC error
            response_body = self.handler.get_raw_error(BadRequestError())
        else:
            response_status = "200 OK"
            response_body = self.handler.handle_raw_request(
                wsgi_input.read(content_length),
                request=self.environ,
            )

        response_headers = [
            ("Content-Type", self._rpc_content_type),
            ("Content-Length", str(len(response_body))),
        ]
        self.start_response(response_status, response_headers)

        yield response_body

    def options(self) -> WSGIResponse:
        response_headers = [
            ("Allow", self._allowed_methods),
            ("Content-Length", "0"),
        ]
        self.start_response("200 OK", response_headers)
        yield b""

    def _not_found(self) -> WSGIResponse:
        response_headers = []
        self.start_response("404 Not Found", response_headers)
        yield b""

    def _method_not_allowed(self) -> WSGIResponse:
        response_headers = [("Allow", self._allowed_methods)]
        self.start_response("405 Method Not Allowed", response_headers)
        yield b""

    def _validate_path(self, path_info: str) -> bool:
        return (
            self.route is not None
            and isinstance(path_info, str)
            and path_info.startswith(self.route)
        )



from __future__ import annotations

from drakaina import remote_procedure
from drakaina.config import rpc_config
from drakaina.contrib.jwt.utils import get_tokens_for
from drakaina.contrib.jwt.utils import get_tokens_by
from drakaina.contrib.jwt.utils import GetNewUserID
from drakaina.contrib.jwt.utils import GetUserID
from drakaina.contrib.jwt.utils import verify_jwt_token
from drakaina.exceptions import AuthenticationFailedError


get_user_id: GetUserID = None
get_new_user_id: GetNewUserID = None


@remote_procedure(name="token", provide_request=True)
def token(request, form_data: dict[str, str]) -> dict[str, str]:
    """todo: docstring"""
    # validate data
    user_id = get_user_id(form_data, request=request)
    if user_id is None:
        raise AuthenticationFailedError(  # todo
            "No active account found with the given credentials",
        )  # "Incorrect email or password"

    # generate tokens
    # RefreshToken, SlidingToken
    tokens = get_tokens_for(user_id)

    return tokens


@remote_procedure(name="register", provide_request=True)
def register(request, form_data: dict[str, str]) -> dict[str, str]:
    """todo: docstring"""
    """
    # Fast API Example
    @app.post("/signup", summary="Create new user", response_model=UserOut)
    async def create_user(data: UserAuth):
        # querying database to check if user already exist
        user = db.get(data.email, None)
        if user is not None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="User with this email already exist",
            )
        user = {
            "email": data.email,
            "password": get_hashed_password(data.password),
            "id": str(uuid4()),
        }
        db[data.email] = user  # saving user to database
        return user

    """
    # validate data
    user_id = get_new_user_id(form_data, request=request)
    if user_id is None:
        raise AuthenticationFailedError(  # todo
            "User with this email already exist",
        )

    # generate tokens
    # RefreshToken, SlidingToken
    tokens = get_tokens_for(user_id)

    return tokens


@remote_procedure(name="token.refresh", provide_request=True)
def refresh_token(refresh: str) -> dict[str, str]:
    """todo: docstring"""
    """
    class TokenRefreshSerializer("serializers.Serializer"):
        Takes a refresh type JSON web token and returns an access type JSON web
        token if the refresh token is valid.

        refresh = serializers.CharField()
        access = serializers.CharField(read_only=True)
        token_class = RefreshToken

        def validate(self, attrs):
            refresh = self.token_class(attrs["refresh"])

            data = {"access": str(refresh.access_token)}

            if api_settings.ROTATE_REFRESH_TOKENS:
                if api_settings.BLACKLIST_AFTER_ROTATION:
                    try:
                        # Attempt to blacklist the given refresh token
                        refresh.blacklist()
                    except AttributeError:
                        # If blacklist app not installed, `blacklist` method will
                        # not be present
                        pass

                refresh.set_jti()
                refresh.set_exp()
                refresh.set_iat()

                data["refresh"] = str(refresh)

            return data

    class TokenRefreshSlidingSerializer("serializers.Serializer"):
        Takes a sliding JSON web token and returns a new, refreshed version if the
        token's refresh period has not expired.

        token = serializers.CharField()
        token_class = "SlidingToken"

        def validate(self, attrs):
            token = self.token_class(attrs["token"])

            # Check that the timestamp in the "refresh_exp" claim has not
            # passed
            token.check_exp("SLIDING_TOKEN_REFRESH_EXP_CLAIM")

            # Update the "exp" and "iat" claims
            token.set_exp()
            token.set_iat()

            return {"token": str(token)}

    class TokenBlacklistSerializer("serializers.Serializer"):
        Takes a token and blacklists it. Must be used with the
        `rest_framework_simplejwt.token_blacklist` app installed.

        refresh = serializers.CharField()
        token_class = "RefreshToken"

        def validate(self, attrs):
            refresh = self.token_class(attrs["refresh"])
            try:
                refresh.blacklist()
            except AttributeError:
                pass
            return {}

    """
    return get_tokens_by(refresh, rpc_config.ROTATE_REFRESH_TOKENS)


@remote_procedure(name="token.verify", provide_request=True)
def verify_token(token: str) -> bool:
    """todo: docstring"""
    verify_jwt_token(token)
    return True
